/******************************************************************************
*       UART单线LED显示接口驱动库
*
* 程序说明：BC759x芯片显示接口的驱动函数库
* 版权所有: 北京凌志比高科技有限公司 https://bitcode.com.cn
* 版本号: V3.2
* 初始版本：2021年3月30日
* 版本历史:
*   2021年3月   V1.0
*   2021年4月   V2.0  修正fifo操作错误
*   2021年4月   V2.1  修正DIRECT_WT拼写错误，重新改写格式以兼容C89等较早C语言标准
*   2021年4月   V3.0  删除缓冲区发送结束时禁止中断操作，以适用于不区分发送/接收中断的设备
*                    修改API函数名称以和BC727X驱动库统一API接口：
*                     uart_tx_ready() ---> tx_ready()
*                     set_tx_eni_func() ---> set_eni_func()
*                     set_tx_disi_func() ---> set_disi_func()
*                     set_uart_wt_func() ---> set_write_func()
*   2021年4月   V3.1  修正display_dec()函数中间0不显示错误
*   2021年5月   V3.2  display_dec()增加显示前导‘0’功能
*   2022年5月   V4.0  增加display_float()函数
*   2023年4月   V4.1  改正了display_dec()函数中的DIG0位显示bug, 增加C89标准的支持
*
* 使用方法：
*   将此头文件加入所有需要使用本驱动库中函数的c源文件，并将led_disp.c文件加入项目源文件列表，
*   即可从用户程序中直接调用本库中函数。有可能需要把key_scan目录加入项目的include搜索目录。
*	详细使用方法，参见《UART单线LED显示接口驱动库技术说明书》
******************************************************************************/
#include "led_disp.h"

const uint8_t BlinkBit[] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };
const uint32_t Mul[8] = {1, 10, 100, 1000, 10000, 100000, 1000000, 10000000};

/* 如果 UART 工作为中断模式，提供FIFO */
#if UART_MODE_INTERRUPT == 1

typedef struct
{
    volatile uint8_t FifoFreeSpace;
    volatile uint8_t Busy;
    uint8_t          NextWritePos;
    volatile uint8_t NextReadPos;
    uint8_t          BlinkCtlL;
    uint8_t          BlinkCtlH;
    uint8_t          UartFifo[UART_FIFO_SIZE];
    void (*pUartWrite)(uint8_t);
    void (*pEnableTxInt)(void);
    void (*pDisableTxInt)(void);
} display_t;

#if __STDC_VERSION__ >= 199901l
display_t Display = {
    .FifoFreeSpace = UART_FIFO_SIZE
};
#else
display_t Display = {UART_FIFO_SIZE};
#endif

/******************************************************************************
* 发送指令
* 这个函数可以用来向BC759x发送任意指令。函数库仅提供了几个上层的常用显示函数，如果用户需要对
* BC759x进行全面的控制，则需要通过调用此函数。上层函数均通过此函数完成BC759x的操作。
* 输入参数:
*     uint8_t Cmd   - 待发送的指令
*     uint8_t Data  - 待发送的数据
* 返回值: 无
******************************************************************************/
void send_cmd(uint8_t Cmd, uint8_t Data)
{
    while (Display.FifoFreeSpace == 0)
        ; /* 检查FIFO中是否还有空间 */
    if (Display.pDisableTxInt != NULL)
    {
        Display.pDisableTxInt(); /* 禁止UART中断防止冲突 */
    }
    Display.UartFifo[Display.NextWritePos] = Cmd;                                               /* 写入新数据 */
    Display.NextWritePos                   = (Display.NextWritePos + 1) & (UART_FIFO_SIZE - 1); /* 调整写指针到下一单元 */
    --Display.FifoFreeSpace;                                                                    /* FIFO空间数-1 */
    if (!Display.Busy)                                                                          /*	如果函数调用之前FIFO已经清空(已经停止自动发送) */
    {
        Display.Busy = 1;
        if (Display.pUartWrite != NULL)
        {
            Display.pUartWrite(Display.UartFifo[Display.NextReadPos]); /*	启动UART发送 */
        }
        Display.NextReadPos = (Display.NextReadPos + 1) & (UART_FIFO_SIZE - 1); /* 调整读指针到下一单元 */
        ++Display.FifoFreeSpace;
    }
    if (Display.pEnableTxInt != NULL)
    {
        Display.pEnableTxInt(); /* 重新使能UART中断 */
    }
    while (Display.FifoFreeSpace == 0) /* 继续将数据字节写入FIFO，首先检查FIFO中是否还有空间 */
        ;
    if (Display.pDisableTxInt != NULL)
    {
        Display.pDisableTxInt(); /* 禁止UART中断防止冲突 */
    }
    --Display.FifoFreeSpace;                                                                    /*	FIFO空间数-1 */
    Display.UartFifo[Display.NextWritePos] = Data;                                              /*	写入新数据 */
    Display.NextWritePos                   = (Display.NextWritePos + 1) & (UART_FIFO_SIZE - 1); /*	调整写指针到下一单元 */
    if (Display.pEnableTxInt != NULL)
    {
        Display.pEnableTxInt(); /* 重新使能UART中断 */
    }
}

/******************************************************************************
* UART发送准备好
* 一般从UART中断中呼叫。数据写入UART后，一般数据发送完成可以继续发送下一数据时，会产生一个
* 中断，此函数由用户在串口中断中呼叫，通知驱动库可以发送FIFO中的下一个数据。
* 输入参数： 无
* 返回值： 无
******************************************************************************/
void tx_ready(void)
{
    if (Display.FifoFreeSpace < UART_FIFO_SIZE) /* 如果FIFO尚未全部发送 */
    {
        if (Display.pUartWrite != NULL)
        {
            Display.pUartWrite(Display.UartFifo[Display.NextReadPos]); /* 发送下一字符 */
        }
        Display.NextReadPos = (Display.NextReadPos + 1) & (UART_FIFO_SIZE - 1);
        ++Display.FifoFreeSpace; /*	可用空间+1 */
    }
    else /* 如果FIFO中已经清空 */
    {
        Display.Busy = 0;
    }
}

/******************************************************************************
* 设置中断使能函数
* 驱动库使用UART中断方式时，需要能够控制中断的使能和禁止，用户需将此操作封装为函数，通过此库函数
* 通知驱动库呼叫哪个函数完成使能串口中断的操作。
* 输入参数： 指向使能串口中断的函数的指针，该函数必须输入和返回值都为void
* 返回值： 无
******************************************************************************/
void set_eni_func(void (*pTxIntEnFunc)(void))
{
    Display.pEnableTxInt = pTxIntEnFunc;
}

/******************************************************************************
* 设置禁止串口中断函数
* 驱动库使用UART中断方式时，需要能够控制中断的使能和禁止，用户需将此操作封装为函数，通过此库函数
* 通知驱动库呼叫哪个函数完成禁止串口中断的操作。
* 输入参数： 指向使能串口中断的函数的指针，该函数必须输入和返回值都为void
* 返回值： 无
******************************************************************************/
void set_disi_func(void (*pTxIntDisFunc)(void))
{
    Display.pDisableTxInt = pTxIntDisFunc;
}

#else /* UART 采用非中断模式 */

struct display_t
{
    uint8_t BlinkCtlL;
    uint8_t BlinkCtlH;
    void (*pUartWrite)(uint8_t);
};

struct display_t Display;

/******************************************************************************
* 发送指令
* 这个函数可以用来向BC759x发送任意指令。函数库仅提供了几个上层的常用显示函数，如果用户需要对
* BC759x进行全面的控制，则需要通过调用此函数。上层函数均通过此函数完成BC759x的操作。
* 输入参数:
*     uint8_t Cmd   - 待发送的指令
*     uint8_t Data  - 待发送的数据
* 返回值: 无
******************************************************************************/
void send_cmd(uint8_t Cmd, uint8_t Data)
{
    Display.pUartWrite(Cmd);
    Display.pUartWrite(Data);
}

#endif

/******************************************************************************
* 设置串口写入函数
* 用户必须将串口的写入操作封装为一个函数，该函数必须以一个uint8_t类型参数为输入，void为返回值。
* 在使用驱动库前，用户必须呼叫此函数通知驱动库哪个函数是串口写入函数。
* 输入参数： 指向串口写入函数的指针，串口写入函数必须以uint8_t为输入参数，void为返回值。
* 返回值： void
******************************************************************************/
void set_write_func(void (*pUartWriteFunc)(uint8_t))
{
    Display.pUartWrite = pUartWriteFunc;
}

/******************************************************************************
* 清除显示
* 这个函数清除所有的显示内容以及将所有的闪烁显示属性归位为不闪烁。
* 输入参数: 无
* 返回值: 无
******************************************************************************/
void clear(void)
{
    uint8_t i;
    send_cmd(WRITE_ALL, 0x00);
    for (i = 0; i < 0x10; i++)
    {
        send_cmd(BLINK_WT_CLR | i, 0xff);
    }
    send_cmd(BLINK_DIG_CTL, 0x00);
    send_cmd(BLINK_DIG_CTL + 1, 0x00);
    Display.BlinkCtlH = 0;
    Display.BlinkCtlL = 0;
}

/******************************************************************************
* 显示十进制数字
* 此函数将输入数值按十进制显示。只接受无符号数值，显示负值需用户自行显示‘-’号。所输入数值，如果
* 高位数字超出了芯片的显示范围，超出部分将不会显示，也不会给出错误信息。如果设置的显示宽度小于
* 实际数字的宽度，超出部分将被忽略，如果超过实际数字的宽度，超出部分将显示”空白“。
* 输入参数:
*     uint32_t Val  - 待显示数值，范围 0 至 4,294,967,295
*     uint8_t Pos   - 个位数字所显示的位置,数字显示的方向，由disp_config.h文件中
*                     LOW_DIG_NUM_ON_RIGHT 决定，该值为1时，个位显示在Pos号显示位，
*                     然后依次使用高标号显示位。
*     uint8_t Width - 低7位表示从最低位开始算起的数值显示的宽度，最高位为1时，不消除前导‘0’
******************************************************************************/
void display_dec(uint32_t Val, uint8_t Pos, uint8_t Width)
{
    if ((Pos > 31) || ((Width & 0x7f) == 0))
    {
        return;
    }
    send_cmd(DECODE_WT | Pos, Val % 10);
    Val = Val / 10;
#if (LOW_DIG_NUM_ON_RIGHT == 1)
    if (++Pos > 31)
    {
        return;
    }
    while ((--Width & 0x7f) > 0)
    {
        if ((Val >= 10) || (Width & 0x80))
        {
            send_cmd(DECODE_WT | Pos, Val % 10);
        }
        else
        {
            send_cmd(DECODE_WT | Pos, (Val % 10) | 0x80);
        }
        Val = Val / 10;
        if (++Pos > 31)
        {
            return;
        }
    }
#else
    if (Pos-- == 0)
    {
        return;
    }
    while ((--Width & 0x7f) > 0)
    {
        if ((Val >= 10) || (Width & 0x80))
        {
            send_cmd(DECODE_WT | Pos, Val % 10);
        }
        else
        {
            send_cmd(DECODE_WT | Pos, (Val % 10) | 0x80);
        }
        Val = Val / 10;
        if (Pos-- == 0)
        {
            return;
        }
    }
#endif
}


/******************************************************************************
* 显示浮点数 display_float()
* 以十进制小数的方式显示浮点数，但不显示小数点和正负号。
* 参数: Val    -- 待显示浮点数
*      Pos    -- 最低位显示位置，即小数点后最低位的位置，使用方法同display_dec()中对应参数
*      Width  -- 从最低位开始算起的数值显示的宽度,包括小数部分和整数部分，其它同display_dec()
*      Frac   -- 小数点后的位数
******************************************************************************/
void display_float(float Val, uint8_t Pos, uint8_t Width, uint8_t Frac)
{
	uint32_t IntVal;
	if (((Width&0x7f) >= Frac) && (Frac <8))
	{
		if (Val < 0)
		{
			Val = -Val;
		}
		IntVal = (uint32_t)(Val*Mul[Frac]+0.5);
		display_dec(IntVal, Pos, Frac|0x80);
#if LOW_DIG_NUM_ON_RIGHT == 1
		display_dec(IntVal/Mul[Frac], Pos+Frac, Width-Frac);
#else
		display_dec(IntVal/Mul[Frac], Pos-Frac, Width-Frac);
#endif
	}
}


/******************************************************************************
* 显示16进制数字
* 此函数将输入数值按16进制显示。此函数最大输入值为0xffff，不过更大数字可以通过拆解和多次调用
* 本函数完成显示。
* 所输入数值，如果高位数字超出了芯片的显示范围，超出部分将不会显示，也不会给出错误信息。
* 输入参数:
*     uint32_t Val  - 待显示数值，范围 0 至 0xffff
*     uint8_t Pos   - 最低位所显示的位置,数字显示的方向，由disp_config.h文件中
*                     LOW_DIG_NUM_ON_RIGHT 决定，该值为1时，最低位显示在Pos号显示位，
*                     然后依次使用高标号显示位。
*     uint8_t Width - 显示宽度。显示宽度小于实际数值宽度时，只显示款度内的数字位；显示宽度
*                     大于实际数字的宽度时，多余位将用‘0’补齐。
******************************************************************************/
void display_hex(uint16_t Val, uint8_t Pos, uint8_t Width)
{
    uint8_t i;
    if (Pos > 31)
    {
        return;
    }
#if (LOW_DIG_NUM_ON_RIGHT == 1)
    if (Width + Pos > 32)
    {
        Width = 32 - Pos;
    }
    for (i = Pos; i < Width + Pos; i++)
    {
        send_cmd(DECODE_WT | i, Val & 0x0f);
        Val >>= 4;
    }
#else
    if (Width > Pos + 1)
    {
        Width = Pos + 1;
    }
    for (i = 0; i < Width; i++)
    {
        send_cmd(DECODE_WT | (Pos - i), Val & 0x0f);
        Val >>= 4;
    }
#endif
}

/******************************************************************************
* 位闪烁控制
* 按数码管位控制闪烁。此函数每次控制一个显示位的闪烁属性，第16位-第31位数码管的闪烁控制如果用户
* 通过直接send_cmd()函数直接改变了状态，再使用本函数对16-31位的闪烁进行控制时，通过直接命令
* 所做的设置可能会丢失。
* 输入参数:
*     uint8_t Digit  - 数码管序号，范围0-31
*     uint8_t OnOff  - 所设置的闪烁状态. 0=不闪烁, 1= 闪烁
* 返回值: 无
******************************************************************************/
void digit_blink(uint8_t Digit, uint8_t OnOff)
{
    if (Digit > 31)
    {
        return;
    }
    if (Digit < 16)
    {
        if (OnOff)
        {
            send_cmd(BLINK_WT_SET | Digit, 0xff);
        }
        else
        {
            send_cmd(BLINK_WT_CLR | Digit, 0xff);
        }
    }
    else
    {
        if (Digit & 0x08)
        {
            Display.BlinkCtlH = OnOff ? Display.BlinkCtlH | BlinkBit[Digit & 0x07] : Display.BlinkCtlH & (~BlinkBit[Digit & 0x07]);
            send_cmd(BLINK_DIG_CTL + 1, Display.BlinkCtlH);
        }
        else
        {
            Display.BlinkCtlL = OnOff ? Display.BlinkCtlL | BlinkBit[Digit & 0x07] : Display.BlinkCtlL & (~BlinkBit[Digit & 0x07]);
            send_cmd(BLINK_DIG_CTL, Display.BlinkCtlL);
        }
    }
}
